#ifdef LOCAL
#include "../../jngen.h"
#else
#include "jngen.h"
#endif

using namespace std;

using ll = long long;

const int MAX_N = 100'000;
const int MAX_VAL = 1'000'000'000 + 7; // MAX_VAL EXCLUSIVE! Must be prime

/**
 * Basic generator. Arguments:
 * -n: Number of integers in expression. Default value is 10^5
 *
 * -lv: Min number in expression
 *
 * -mv: Max number in expression(excluded). Default is MAX_VAL
 *
 * -sm: How to split blocks. Described with string pattern:
 *      digit - small block with this length
 *      'X' - large block. All X blocks will have ~equal len
 *
 *      If pattern is empty, something will be used instead
 *
 * -cp: Digit change policy.
 *      0 - Randomly change digits. Default value
 *      1 - Change one digit in two numbers from the same multiply block.
 *          If there is no such block, digits will be changed randomly
 *      2 - Change digits in different multiply blocks.
 *          If there is the only block, change randomly
 *      3 - Change two digits in one number.
 *          If all numbers have one digit, change randomly
 *      4 - Change one digit
 *      5 - Change three digits
 *      6 - Do stuff so answer will probably be NO
 * */

int n;
int lv;
int mv;
int cp;
string split_mode;

ll get_number_len(ll x) {
  int len = 1;
  ll bound = 10;
  while (x >= bound) {
    len++;
    bound *= 10;
  }
  return len;
}

ll set_digit(ll x, ll p10, ll new_val) {
  ll old = (x / p10) % 10;
  return x + (new_val - old) * p10;
}

vector<int> split_n() {
  if (split_mode.empty()) {
    split_mode = rnd.next("X[1-9]{5}");
  }
  vector<int> spl;
  int rest = n, zeroes = 0;
  for (char c : split_mode) {
    if (c == 'X') {
      zeroes++;
      spl.push_back(0);
    } else {
      spl.push_back(c - '0');
      rest -= c - '0';
    }
  }
  if (zeroes == 0) {
    zeroes = 1;
    spl.insert(spl.begin(), 0);
  }
  rest /= zeroes;
  for (int &x : spl) {
    if (x == 0) {
      x = rest;
    }
  }
  return spl;
}

vector<vector<ll>> get_blocks() {
  auto split = split_n();
  vector<vector<ll>> result(split.size());
  for (int i = 0; i < split.size(); i++) {
    result[i] = Array64::random(split[i], lv, mv);
  }
  return result;
}

ll eval(const vector<vector<ll>> &blocks) {
  ll res = 0;
  for (const vector<ll> &v : blocks) {
    ll prod = 1;
    for (ll x : v) {
      prod = (prod * x) % MAX_VAL;
    }
    res = (res + prod) % MAX_VAL;
  }
  return res;
}

void print_test(const vector<vector<ll>> &blocks, ll value) {
  for (int i = 0; i < blocks.size(); i++) {
    for (int j = 0; j < blocks[i].size(); j++) {
      cout << blocks[i][j] << " ";
      if (j + 1 != blocks[i].size()) cout << "* ";
    }
    if (i + 1 != blocks.size()) {
      cout << "+ ";
    } else {
      cout << "= ";
    }
  }
  cout << value << "\n";
}

void change_number(ll &number) {
  static ll p10[10] = {1, 10, 100, 1'000, 10'000,
                       100'000, 1'000'000, 10'000'000, 100'000'000, 1'000'000'000};
  ll len = get_number_len(number);
  vector<ll> choices;
  for (int i = 0; i < len; i++) {
    for (int d = ((i + 1 < len or len == 1) ? 0 : 1); d < 10; d++) {
      ll new_val = set_digit(number, p10[i], d);
      if (new_val < MAX_VAL) {
        choices.push_back(new_val);
      }
    }
  }
  number = rnd.choice(choices);
}

pair<int, int> get_rnd_pos(const vector<vector<ll>> &bl) {
  int i = rnd.next(bl.size());
  int j = rnd.next(bl[i].size());
  return {i ,j};
}

void c0(vector<vector<ll>> &blocks) {
  auto p1 = get_rnd_pos(blocks);
  auto p2 = get_rnd_pos(blocks);
  change_number(blocks[p1.first][p1.second]);
  change_number(blocks[p2.first][p2.second]);
}

void c1(vector<vector<ll>> &blocks) {
  vector<int> indexes;
  for (int i = 0; i < blocks.size(); i++) {
    if (blocks[i].size() > 1) {
      indexes.push_back(i);
    }
  }
  if (indexes.empty()) {
    c0(blocks);
  } else {
    int i = rnd.choice(indexes);
    auto p = rnd.nextp(blocks[i].size(), odpair);
    change_number(blocks[i][p.first]);
    change_number(blocks[i][p.second]);
  }
}

void c2(vector<vector<ll>> &blocks) {
  if (blocks.size() == 1) {
    c0(blocks);
  } else {
    auto pi = rnd.nextp(blocks.size(), odpair);
    int j1 = rnd.next(blocks[pi.first].size());
    int j2 = rnd.next(blocks[pi.second].size());
    change_number(blocks[pi.first][j1]);
    change_number(blocks[pi.second][j2]);
  }
}

void c3(vector<vector<ll>> &blocks) {
  vector<pair<int, int>> long_enough;
  for (int i = 0; i < blocks.size(); i++) {
    for (int j = 0; j < blocks[i].size(); j++) {
      if (blocks[i][j] > 9) {
        long_enough.emplace_back(i, j);
      }
    }
  }
  if (long_enough.empty()) {
    c0(blocks);
  } else {
    auto p = rnd.choice(long_enough);
    change_number(blocks[p.first][p.second]);
    change_number(blocks[p.first][p.second]);
  }
}

void c4(vector<vector<ll>> &blocks) {
  auto p = get_rnd_pos(blocks);
  change_number(blocks[p.first][p.second]);
}

void c5(vector<vector<ll>> &blocks) {
  c4(blocks);
  c4(blocks);
  c4(blocks);
}

void c6(vector<vector<ll>> &blocks) {
  int steps = rnd.next(3, max(n, 20));
  for (int i = 0; i < steps; i++) {
    c4(blocks);
  }
}

vector<void(*)(vector<vector<ll>>&)> modifiers =
    { c0, c1, c2, c3, c4, c5, c6 };

int main(int argc, char *argv[]) {
  parseArgs(argc, argv);
  registerGen(argc, argv, 1);
  n = getOpt("n", MAX_N);
  lv = getOpt("lv", 0);
  mv = getOpt("mv", MAX_VAL);
  mv--; // mv is inclusive now due to lv usage
  cp = getOpt("cp", 0);
  split_mode = getOpt("sm", "");

  auto expression = get_blocks();
  ll result = eval(expression);

  //print_test(expression, result);

  modifiers[cp](expression);
  print_test(expression, result);
}