#ifdef LOCAL
#include "../../jngen.h"
#else
#include "jngen.h"
#endif

using namespace std;

using ll = long long;

const int MAX_N = 100'000;
const int MAX_VAL = 1'000'000'000 + 7; // MAX_VAL EXCLUSIVE! Must be prime

/**
 * Basic generator. Arguments:
 * -n: Number of integers in expression. Default value is 10^5
 *
 * -mv: Max number in expression. Default is MAX_VAL
 *
 * -dl, dr: Length range for multiply blocks.
 *          Defaults are 1 and -1(so any size may be)
 *
 * -blk: Number of blocks. Default value is ~sqrt(n)
 *
 * -sb - small numbers bound(exclusive)
 *
 * -sn - number of small nums
 *
 * -z - number of zeroes. if greater than sn, than it is chosen randomly
 *
 * -cp: Digit change policy.
 *      0 - Randomly change digits. Default value
 *      1 - Change one digit in two numbers from the same multiply block.
 *          If there is no such block, digits will be changed randomly
 *      2 - Change digits in different multiply blocks.
 *          If there is the only block, change randomly
 *      3 - Change two digits in one number.
 *          If all numbers have one digit, change randomly
 *      4 - Change one digit
 *      5 - Change three digits
 *      6 - Do stuff so answer will probably be NO
 *
 *      if cp is >= 10, than cp - 10 policy is applied to small numbers
 * */

int n;
int lv;
int mv;
int dl;
int dr;
int blk;
int cp;
int small_bound;
int small_number;
int zeroes;

ll get_number_len(ll x) {
  int len = 1;
  ll bound = 10;
  while (x >= bound) {
    len++;
    bound *= 10;
  }
  return len;
}

ll set_digit(ll x, ll p10, ll new_val) {
  ll old = (x / p10) % 10;
  return x + (new_val - old) * p10;
}

vector<vector<ll>> get_blocks() {
  auto split = jngen::MathRandom::partition(n, blk, 1, n);
  vector<vector<ll>> result(blk);
  for (int i = 0; i < blk; i++) {
    result[i] = Array64::random(split[i], lv, mv);
  }
  return result;
}

ll eval(const vector<vector<ll>> &blocks) {
  ll res = 0;
  for (const vector<ll> &v : blocks) {
    ll prod = 1;
    for (ll x : v) {
      prod = (prod * x) % MAX_VAL;
    }
    res = (res + prod) % MAX_VAL;
  }
  return res;
}

void print_test(const vector<vector<ll>> &blocks, ll value) {
  for (int i = 0; i < blocks.size(); i++) {
    for (int j = 0; j < blocks[i].size(); j++) {
      cout << blocks[i][j] << " ";
      if (j + 1 != blocks[i].size()) cout << "* ";
    }
    if (i + 1 != blocks.size()) {
      cout << "+ ";
    } else {
      cout << "= ";
    }
  }
  cout << value << "\n";
}

void change_number(ll &number) {
  static ll p10[10] = {1, 10, 100, 1'000, 10'000,
                       100'000, 1'000'000, 10'000'000, 100'000'000, 1'000'000'000};
  ll len = get_number_len(number);
  vector<ll> choices;
  for (int i = 0; i < len; i++) {
    for (int d = ((i + 1 < len or len == 1) ? 0 : 1); d < 10; d++) {
      ll new_val = set_digit(number, p10[i], d);
      if (new_val < MAX_VAL) {
        choices.push_back(new_val);
      }
    }
  }
  number = rnd.choice(choices);
}

function<pair<int, int>(const vector<vector<ll>>&)> get_rnd_pos = [](const vector<vector<ll>> &bl) {
  int i = rnd.next(bl.size());
  int j = rnd.next(bl[i].size());
  return make_pair(i, j);
};

void c0(vector<vector<ll>> &blocks) {
  auto p1 = get_rnd_pos(blocks);
  auto p2 = get_rnd_pos(blocks);
  change_number(blocks[p1.first][p1.second]);
  change_number(blocks[p2.first][p2.second]);
}

void c1(vector<vector<ll>> &blocks) {
  vector<int> indexes;
  for (int i = 0; i < blocks.size(); i++) {
    if (blocks[i].size() > 1) {
      indexes.push_back(i);
    }
  }
  if (indexes.empty()) {
    c0(blocks);
  } else {
    int i = rnd.choice(indexes);
    auto p = rnd.nextp(blocks[i].size(), odpair);
    change_number(blocks[i][p.first]);
    change_number(blocks[i][p.second]);
  }
}

void c2(vector<vector<ll>> &blocks) {
  if (blocks.size() == 1) {
    c0(blocks);
  } else {
    auto pi = rnd.nextp(blocks.size(), odpair);
    int j1 = rnd.next(blocks[pi.first].size());
    int j2 = rnd.next(blocks[pi.second].size());
    change_number(blocks[pi.first][j1]);
    change_number(blocks[pi.second][j2]);
  }
}

void c3(vector<vector<ll>> &blocks) {
  vector<pair<int, int>> long_enough;
  for (int i = 0; i < blocks.size(); i++) {
    for (int j = 0; j < blocks[i].size(); j++) {
      if (blocks[i][j] > 9) {
        long_enough.emplace_back(i, j);
      }
    }
  }
  if (long_enough.empty()) {
    c0(blocks);
  } else {
    auto p = rnd.choice(long_enough);
    change_number(blocks[p.first][p.second]);
    change_number(blocks[p.first][p.second]);
  }
}

void c4(vector<vector<ll>> &blocks) {
  auto p = get_rnd_pos(blocks);
  change_number(blocks[p.first][p.second]);
}

void c5(vector<vector<ll>> &blocks) {
  c4(blocks);
  c4(blocks);
  c4(blocks);
}

void c6(vector<vector<ll>> &blocks) {
  int steps = rnd.next(3, max(n, 20));
  for (int i = 0; i < steps; i++) {
    c4(blocks);
  }
}

vector<void(*)(vector<vector<ll>>&)> modifiers =
    { c0, c1, c2, c3, c4, c5, c6 };

int main(int argc, char *argv[]) {
  parseArgs(argc, argv);
  registerGen(argc, argv, 1);
  n = getOpt("n", MAX_N);
  lv = getOpt("lv", 0);
  mv = getOpt("mv", MAX_VAL);
  mv--;

  blk = getOpt("blk", min(n, (int)sqrt(n) + 1));
  cp = getOpt("cp", 0);
  small_bound = getOpt("sb", 10);
  small_number = getOpt("sn", 0);
  zeroes = getOpt("z", 0);
  if (zeroes > small_number) {
    zeroes = rnd.next(0, small_number);
  }

  auto expression = get_blocks();
  vector<pair<int, int>> changed_small;

  for (int i = 0; i < small_number; i++) {
    auto pos = get_rnd_pos(expression);
    changed_small.push_back(pos);
    ll val = (i < zeroes) ? 0 : rnd.next(1, small_bound - 1);
    expression[pos.first][pos.second] = val;
  }
  ll result = eval(expression);

  //print_test(expression, result);

  if (cp >= 10) {
    get_rnd_pos = [&changed_small](const vector<vector<ll>>&) {
      return rnd.choice(changed_small);
    };
    cp -= 10;
  }
  modifiers[cp](expression);
  print_test(expression, result);
}